@@ -20,7 +20,7 @@
 
 #if (ZFEX_ARM_NEON_FEATURE == 1)
 #include <arm_neon.h>
-#endif /* ZFEX_ARM_NEON_FEATURE */
+#endif /* ZFEX_ARM_NEON_FEATURE == 1 */
 
 /*
  * Primitive polynomials - see Lin & Costello, Appendix A,
@@ -56,8 +56,6 @@ modnn(int x) {
     return x;
 }
 
-#define SWAP(a,b,t) {t tmp; tmp=a; a=b; b=tmp;}
-
 /*
  * gf_mul(x,y) multiplies two numbers.  It is much faster to use a
  * multiplication table.
@@ -191,6 +189,73 @@ generate_gf (void) {
  * Various linear algebra operations that i use often.
  */
 
+#if (ZFEX_ARM_NEON_FEATURE == 1)
+static inline
+void addmul_neon_kernel(
+    uint8_t *dst,
+    uint8_t const *src,
+    uint8x16_t const vmul_lo,
+    uint8x16_t const vmul_hi,
+    uint8x16_t const mask_0F)
+{
+    dst = ZFEX_ASSUME_ALIGNED(dst, ZFEX_SIMD_ALIGNMENT);
+    src = ZFEX_ASSUME_ALIGNED(src, ZFEX_SIMD_ALIGNMENT);
+
+    register uint8x16_t q2  = vld1q_u8(src);
+    register uint8x16_t q1  = vld1q_u8(dst);
+
+    register uint8x16_t q9  = q2 & mask_0F;
+    register uint8x16_t q10  = q2 >> 4;
+
+    __asm__ ("vtbl.8 %e[x], {%q[t]}, %e[x]" : [x]"+w"(q9) : [t]"w"(vmul_lo));
+    __asm__ ("vtbl.8 %f[x], {%q[t]}, %f[x]" : [x]"+w"(q9) : [t]"w"(vmul_lo));
+    __asm__ ("vtbl.8 %e[x], {%q[t]}, %e[x]" : [x]"+w"(q10) : [t]"w"(vmul_hi));
+    __asm__ ("vtbl.8 %f[x], {%q[t]}, %f[x]" : [x]"+w"(q10) : [t]"w"(vmul_hi));
+
+    vst1q_u8(dst, q1 ^ q9 ^ q10);
+}
+#endif /* (ZFEX_ARM_NEON_FEATURE == 1) */
+
+#if (ZFEX_INTEL_SSSE3_FEATURE == 1)
+static inline
+void addmul_ssse3_kernel_aligned(
+    uint8_t *dst,
+    uint8_t const *src,
+    __m128i const vmul_lo,
+    __m128i const vmul_hi,
+    __v16qu const mask_0F)
+{
+    dst = ZFEX_ASSUME_ALIGNED(dst, ZFEX_SIMD_ALIGNMENT);
+    src = ZFEX_ASSUME_ALIGNED(src, ZFEX_SIMD_ALIGNMENT);
+
+    register __v16qu const vsrc = (__v16qu)_mm_load_si128((__m128i const *)src);
+    register __v16qu const vsrc_lo = vsrc & mask_0F;
+    register __v16qu const vsrc_hi = (__v16qu)((__v8hu)vsrc >> 4) & mask_0F;
+
+    register __m128i const to_xor = _mm_shuffle_epi8(vmul_lo, (__m128i)vsrc_lo) ^ _mm_shuffle_epi8(vmul_hi, (__m128i)vsrc_hi);
+
+    _mm_store_si128((__m128i *)dst, to_xor ^ _mm_load_si128((__m128i const *)dst));
+}
+
+
+static inline
+void addmul_ssse3_kernel_unaligned(
+    uint8_t *dst,
+    uint8_t const *src,
+    __m128i const vmul_lo,
+    __m128i const vmul_hi,
+    __v16qu const mask_0F)
+{
+    register __v16qu const vsrc = (__v16qu)_mm_lddqu_si128((__m128i const *)src);
+    register __v16qu const vsrc_lo = vsrc & mask_0F;
+    register __v16qu const vsrc_hi = (__v16qu)((__v8hu)vsrc >> 4) & mask_0F;
+
+    register __m128i const to_xor = _mm_shuffle_epi8(vmul_lo, (__m128i)vsrc_lo) ^ _mm_shuffle_epi8(vmul_hi, (__m128i)vsrc_hi);
+
+    _mm_storeu_si128((__m128i *)dst, to_xor ^ _mm_lddqu_si128((__m128i const *)dst));
+}
+#endif /* (ZFEX_INTEL_SSSE3_FEATURE == 1) */
+
 /*
  * addmul() computes dst[] = dst[] + c * src[]
  * This is used often, so better optimize it! Currently the loop is
@@ -201,32 +266,79 @@ generate_gf (void) {
 #define addmul(dst, src, c, sz)                 \
     if (c != 0) _addmul1(dst, src, c, sz)
 
-#ifndef UNROLL
-#define UNROLL 16               /* 1, 4, 8, 16 */
-#endif
 static
 #if (ZFEX_INLINE_ADDMUL_FEATURE == 1)
 inline
 #endif /* ZFEX_INLINE_ADDMUL_FEATURE */
-void _addmul1(register gf*restrict dst, register const gf*restrict src, gf c, size_t sz) {
+void _addmul1(register gf* ZFEX_RESTRICT dst, register const gf* ZFEX_RESTRICT src, gf c, size_t sz)
+{
+#if (ZFEX_INTEL_SSSE3_FEATURE == 1)
+    enum { ZFEX_UNROLL_ADDMUL_UNIT = sizeof (__m128i) };
+    enum { ZFEX_UNROLL_ADDMUL_TILE_1 = ZFEX_UNROLL_ADDMUL_UNIT };
+    enum { ZFEX_UNROLL_ADDMUL_TILE = ZFEX_UNROLL_ADDMUL_UNIT * (ZFEX_UNROLL_ADDMUL_SIMD) };
+
+    const gf* lim = &dst[sz];
+
+    register __m128i const vmul_lo = _mm_load_si128((__m128i const *)gf_mul_table[c]);
+    register __m128i const vmul_hi = _mm_load_si128((__m128i const *)gf_mul_table_16[c]);
+
+    register __v16qu const mask0F = (__v16qu)_mm_set1_epi8(0x0F);
+
+#if (ZFEX_UNROLL_ADDMUL_SIMD > 1)
+    lim -= ZFEX_UNROLL_ADDMUL_TILE - 1;
+    for (; dst < lim; dst += ZFEX_UNROLL_ADDMUL_TILE, src += ZFEX_UNROLL_ADDMUL_TILE)
+    {
+#define KERNEL(i) addmul_ssse3_kernel_unaligned(dst + i * ZFEX_UNROLL_ADDMUL_UNIT, src + i * ZFEX_UNROLL_ADDMUL_UNIT, vmul_lo, vmul_hi, mask0F);
+        PP_REPEAT(ZFEX_UNROLL_ADDMUL_SIMD, KERNEL)
+#undef KERNEL
+    }
+    lim += ZFEX_UNROLL_ADDMUL_TILE - 1;
+#endif /* (ZFEX_UNROLL_ADDMUL_SIMD > 1) */
+
+    lim -= ZFEX_UNROLL_ADDMUL_TILE_1 - 1;
+    for (; dst < lim; dst += ZFEX_UNROLL_ADDMUL_TILE_1, src += ZFEX_UNROLL_ADDMUL_TILE_1)
+    {
+        addmul_ssse3_kernel_unaligned(dst, src, vmul_lo, vmul_hi, mask0F);
+    }
+    lim += ZFEX_UNROLL_ADDMUL_TILE_1 - 1;
+
+    if (dst < lim)
+    {
+        register __v16qu const vsrc = (__v16qu)_mm_lddqu_si128((__m128i const *)src);
+        register __v16qu const vsrc_lo = vsrc & mask0F;
+        register __v16qu const vsrc_hi = (__v16qu)((__v8hu)vsrc >> 4) & mask0F;
+
+        register __m128i const tail_mask = mask_to_u128_SSSE3(0xFFFF >> (16 - (lim - dst)));
+
+        register __m128i const to_xor = (_mm_shuffle_epi8(vmul_lo, (__m128i)vsrc_lo) ^ _mm_shuffle_epi8(vmul_hi, (__m128i)vsrc_hi));
+
+        _mm_maskmoveu_si128(to_xor ^ _mm_lddqu_si128((__m128i const *)dst), tail_mask, (char *)dst);
+    }
+
+#else /* not ZFEX_INTEL_SSSE3_FEATURE */
+    enum { ZFEX_UNROLL_ADDMUL_UNIT = 1 };
+    enum { ZFEX_UNROLL_ADDMUL_TILE = ZFEX_UNROLL_ADDMUL_UNIT * (ZFEX_UNROLL_ADDMUL) };
+
     USE_GF_MULC;
-    const gf* lim = &dst[sz - UNROLL + 1];
+    const gf* lim = &dst[sz - ZFEX_UNROLL_ADDMUL_TILE + 1];
 
     GF_MULC0 (c);
 
-
-#if (UNROLL > 1)                /* unrolling by 8/16 is quite effective on the pentium */
-    for (; dst < lim; dst += UNROLL, src += UNROLL) {
-#define OP(i) GF_ADDMULC (dst[i], src[i]);
-        PP_REPEAT(UNROLL, OP)
-#undef OP
+#if (ZFEX_UNROLL_ADDMUL > 1)
+    for (; dst < lim; dst += ZFEX_UNROLL_ADDMUL_TILE, src += ZFEX_UNROLL_ADDMUL_TILE)
+    {
+#define KERNEL(i) GF_ADDMULC (dst[i], src[i]);
+        PP_REPEAT(ZFEX_UNROLL_ADDMUL, KERNEL)
+#undef KERNEL
     }
 #endif
 
-
-    lim += UNROLL - 1;
+    lim += ZFEX_UNROLL_ADDMUL_TILE - 1;
     for (; dst < lim; dst++, src++)       /* final components */
+    {
         GF_ADDMULC (*dst, *src);
+    }
+#endif
 }
 
 #define addmul_simd(dst, src, c, sz)                 \
@@ -236,32 +348,40 @@ static
 #if (ZFEX_INLINE_ADDMUL_SIMD_FEATURE == 1)
 inline
 #endif /* ZFEX_INLINE_ADDMUL_SIMD_FEATURE */
-void _addmul1_simd(register gf * restrict dst, register const gf * restrict src, gf c, size_t sz)
+void _addmul1_simd(register gf * ZFEX_RESTRICT dst, register const gf * ZFEX_RESTRICT src, gf c, size_t sz)
 {
     dst = ZFEX_ASSUME_ALIGNED(dst, ZFEX_SIMD_ALIGNMENT);
     src = ZFEX_ASSUME_ALIGNED(src, ZFEX_SIMD_ALIGNMENT);
 
-#if (ZFEX_INTEL_SSSE3_FEATURE == 1) && (UNROLL == 16)
+#if (ZFEX_INTEL_SSSE3_FEATURE == 1)
+    enum { ZFEX_UNROLL_ADDMUL_UNIT = sizeof (__m128i) };
+    enum { ZFEX_UNROLL_ADDMUL_TILE_1 = ZFEX_UNROLL_ADDMUL_UNIT };
+    enum { ZFEX_UNROLL_ADDMUL_TILE = ZFEX_UNROLL_ADDMUL_UNIT * (ZFEX_UNROLL_ADDMUL_SIMD) };
 
-    const gf* lim = &dst[sz - UNROLL + 1];
+    const gf* lim = &dst[sz];
 
     register __m128i const vmul_lo = _mm_load_si128((__m128i const *)gf_mul_table[c]);
     register __m128i const vmul_hi = _mm_load_si128((__m128i const *)gf_mul_table_16[c]);
 
     register __v16qu const mask0F = (__v16qu)_mm_set1_epi8(0x0F);
 
-    for (; dst < lim; dst += UNROLL, src += UNROLL)
+#if (ZFEX_UNROLL_ADDMUL_SIMD > 1)
+    lim -= ZFEX_UNROLL_ADDMUL_TILE - 1;
+    for (; dst < lim; dst += ZFEX_UNROLL_ADDMUL_TILE, src += ZFEX_UNROLL_ADDMUL_TILE)
     {
-        register __v16qu const vsrc = (__v16qu)_mm_load_si128((__m128i const *)src);
-        register __v16qu const vsrc_lo = vsrc & mask0F;
-        register __v16qu const vsrc_hi = (__v16qu)((__v8hu)vsrc >> 4) & mask0F;
-
-        register __m128i const to_xor = _mm_shuffle_epi8(vmul_lo, (__m128i)vsrc_lo) ^ _mm_shuffle_epi8(vmul_hi, (__m128i)vsrc_hi);
-
-        _mm_store_si128((__m128i *)dst, to_xor ^ _mm_load_si128((__m128i const *)dst));
+#define KERNEL(i) addmul_ssse3_kernel_aligned(dst + i * ZFEX_UNROLL_ADDMUL_UNIT, src + i * ZFEX_UNROLL_ADDMUL_UNIT, vmul_lo, vmul_hi, mask0F);
+        PP_REPEAT(ZFEX_UNROLL_ADDMUL_SIMD, KERNEL)
+#undef KERNEL
     }
+    lim += ZFEX_UNROLL_ADDMUL_TILE - 1;
+#endif /* (ZFEX_UNROLL_ADDMUL_SIMD > 1) */
 
-    lim += UNROLL - 1;
+    lim -= ZFEX_UNROLL_ADDMUL_TILE_1 - 1;
+    for (; dst < lim; dst += ZFEX_UNROLL_ADDMUL_TILE_1, src += ZFEX_UNROLL_ADDMUL_TILE_1)
+    {
+        addmul_ssse3_kernel_aligned(dst, src, vmul_lo, vmul_hi, mask0F);
+    }
+    lim += ZFEX_UNROLL_ADDMUL_TILE_1 - 1;
 
     if (dst < lim)
     {
@@ -271,75 +391,100 @@ void _addmul1_simd(register gf * restrict dst, register const gf * restrict src,
 
         register __m128i const tail_mask = mask_to_u128_SSSE3(0xFFFF >> (16 - (lim - dst)));
 
-        register __m128i const to_xor = (_mm_shuffle_epi8(vmul_lo, (__m128i)vsrc_lo) ^ _mm_shuffle_epi8(vmul_hi, (__m128i)vsrc_hi))
-            & tail_mask;
+        register __m128i const to_xor = (_mm_shuffle_epi8(vmul_lo, (__m128i)vsrc_lo) ^ _mm_shuffle_epi8(vmul_hi, (__m128i)vsrc_hi));
 
-        _mm_store_si128((__m128i *)dst, to_xor ^ _mm_load_si128((__m128i const *)dst));
+        _mm_maskmoveu_si128(to_xor ^ _mm_load_si128((__m128i const *)dst), tail_mask, (char *)dst);
     }
 
-#elif (ZFEX_ARM_NEON_FEATURE == 1) && (UNROLL == 16)
-    // Idea from https://botan.randombit.net zfec_vperm.cpp
-    const gf* lim = &dst[sz - UNROLL + 1];
+#elif (ZFEX_ARM_NEON_FEATURE == 1)
+    enum { ZFEX_UNROLL_ADDMUL_UNIT = sizeof (uint8x16_t) };
+    enum { ZFEX_UNROLL_ADDMUL_TILE_1 = ZFEX_UNROLL_ADDMUL_UNIT };
+    enum { ZFEX_UNROLL_ADDMUL_TILE = ZFEX_UNROLL_ADDMUL_UNIT * (ZFEX_UNROLL_ADDMUL_SIMD) };
+
+    const gf* lim = &dst[sz];
 
     register uint8x16_t q0  = {0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F};
     register uint8x16_t q3  = vld1q_u8(gf_mul_table[c]);
     register uint8x16_t q8  = vld1q_u8(gf_mul_table_16[c]);
 
-    for (; dst < lim; dst += UNROLL, src += UNROLL)
+#if (ZFEX_UNROLL_ADDMUL_SIMD > 1)
+    lim -= ZFEX_UNROLL_ADDMUL_TILE - 1;
+    for (; dst < lim; dst += ZFEX_UNROLL_ADDMUL_TILE, src += ZFEX_UNROLL_ADDMUL_TILE)
     {
-         register uint8x16_t q2  = vld1q_u8(src);
-         register uint8x16_t q1  = vld1q_u8(dst);
-
-         register uint8x16_t q9  = q2 & q0;
-         register uint8x16_t q10  = q2 >> 4;
-
-        __asm__ ("vtbl.8 %e[x], {%q[t]}, %e[x]" : [x]"+w"(q9) : [t]"w"(q3));
-        __asm__ ("vtbl.8 %f[x], {%q[t]}, %f[x]" : [x]"+w"(q9) : [t]"w"(q3));
-        __asm__ ("vtbl.8 %e[x], {%q[t]}, %e[x]" : [x]"+w"(q10) : [t]"w"(q8));
-        __asm__ ("vtbl.8 %f[x], {%q[t]}, %f[x]" : [x]"+w"(q10) : [t]"w"(q8));
-
-        vst1q_u8(dst, q1 ^ q9 ^ q10);
+#define KERNEL(i) addmul_neon_kernel(dst + i * ZFEX_UNROLL_ADDMUL_UNIT, src + i * ZFEX_UNROLL_ADDMUL_UNIT, q3, q8, q0);
+        PP_REPEAT(ZFEX_UNROLL_ADDMUL_SIMD, KERNEL)
+#undef KERNEL
     }
+    lim += ZFEX_UNROLL_ADDMUL_TILE - 1;
+#endif /* (ZFEX_UNROLL_ADDMUL_SIMD > 1) */
 
-    lim += UNROLL - 1;
+    lim -= ZFEX_UNROLL_ADDMUL_TILE_1 - 1;
+    for (; dst < lim; dst += ZFEX_UNROLL_ADDMUL_TILE_1, src += ZFEX_UNROLL_ADDMUL_TILE_1)
+    {
+        addmul_neon_kernel(dst, src, q3, q8, q0);
+    }
+    lim += ZFEX_UNROLL_ADDMUL_TILE_1 - 1;
 
     if (dst < lim)
     {
-         register uint8x16_t q2  = vld1q_u8(src);
-         register uint8x16_t q1  = vld1q_u8(dst);
+        register uint8x16_t q2 = vld1q_u8(src);
+        register uint8x16_t q1 = vld1q_u8(dst);
 
-         register uint8x16_t q9  = q2 & q0;
-         register uint8x16_t q10  = q2 >> 4;
+        register uint8x16_t q9 = q2 & q0;
+        register uint8x16_t q10 = q2 >> 4;
 
         __asm__ ("vtbl.8 %e[x], {%q[t]}, %e[x]" : [x]"+w"(q9) : [t]"w"(q3));
         __asm__ ("vtbl.8 %f[x], {%q[t]}, %f[x]" : [x]"+w"(q9) : [t]"w"(q3));
         __asm__ ("vtbl.8 %e[x], {%q[t]}, %e[x]" : [x]"+w"(q10) : [t]"w"(q8));
         __asm__ ("vtbl.8 %f[x], {%q[t]}, %f[x]" : [x]"+w"(q10) : [t]"w"(q8));
 
-        register uint8x16_t const tail_mask = mask_to_u128_NEON(0xFFFF >> (16 - (lim - dst)));
+#if (ZFEX_IS_LITTLE_ENDIAN == 1)
+        uint16_t const bitmask = 0xFFFF >> (16 - (lim - dst));
+        register uint8x16_t const tail_mask = mask_to_u128_NEON(bitmask);
+#else
+        /*
+         * On big endian ARM we need to take care of two things:
+         * 1. Loads with vld1q are always executed into a pair of dN registers.
+         *    Their order is always the same and endianness only matters within
+         *    each of them. As a result, we need to swap higher and lower part
+         *    of the mask (since the bitmask is 16-bit byteswap is needed).
+         * 2. Endianness means that to generate the mask we need to shift in
+         *    the opposite direction.
+         * TODO: at some point it would be interesting checking how does simple
+         * LUT would perform to get the bytemask we need.
+         */
+        uint16_t const bitmask = 0xFFFF << (16 - (lim - dst));
+        register uint8x16_t const tail_mask = mask_to_u128_NEON(
+            (bitmask << 8) | (bitmask >> 8)
+        );
+#endif
 
         vst1q_u8(dst, q1 ^ ((q9 ^ q10) & tail_mask));
     }
 
 #else /* not ZFEX_INTEL_SSSE3_FEATURE && not ZFEX_ARM_NEON_FEATURE */
+    enum { ZFEX_UNROLL_ADDMUL_UNIT = 1 };
+    enum { ZFEX_UNROLL_ADDMUL_TILE = ZFEX_UNROLL_ADDMUL_UNIT * (ZFEX_UNROLL_ADDMUL) };
 
     USE_GF_MULC;
-    const gf* lim = &dst[sz - UNROLL + 1];
+    const gf* lim = &dst[sz - ZFEX_UNROLL_ADDMUL_TILE + 1];
 
     GF_MULC0 (c);
 
-
-#if (UNROLL > 1)
-    for (; dst < lim; dst += UNROLL, src += UNROLL) {
-#define OP(i) GF_ADDMULC (dst[i], src[i]);
-        PP_REPEAT(UNROLL, OP)
-#undef OP
+#if (ZFEX_UNROLL_ADDMUL > 1)
+    for (; dst < lim; dst += ZFEX_UNROLL_ADDMUL_TILE, src += ZFEX_UNROLL_ADDMUL_TILE)
+    {
+#define KERNEL(i) GF_ADDMULC (dst[i], src[i]);
+        PP_REPEAT(ZFEX_UNROLL_ADDMUL, KERNEL)
+#undef KERNEL
     }
 #endif
 
-    lim += UNROLL - 1;
+    lim += ZFEX_UNROLL_ADDMUL_TILE - 1;
     for (; dst < lim; dst++, src++)       /* final components */
+    {
         GF_ADDMULC (*dst, *src);
+    }
 #endif
 }
 
@@ -369,7 +514,10 @@ _matmul(gf * a, gf * b, gf * c, unsigned n, unsigned k, unsigned m) {
  * Return non-zero if singular.
  */
 static void
-_invert_mat(gf* src, size_t k) {
+_invert_mat(gf* src, size_t k)
+{
+#define SWAP(a, b, Tp) {Tp t = a; a = b; b = t;}
+
     gf c;
     size_t irow = 0;
     size_t icol = 0;
@@ -465,6 +613,7 @@ _invert_mat(gf* src, size_t k) {
     free(indxr);
     free(ipiv);
     free(id_row);
+#undef SWAP
 }
 
 /*
@@ -478,7 +627,7 @@ _invert_mat(gf* src, size_t k) {
  * p = coefficients of the matrix (p_i)
  * q = values of the polynomial (known)
  */
-void
+static void
 _invert_vdm (gf* src, unsigned k) {
     unsigned i, j, row, col;
     gf *b, *c, *p;
@@ -547,17 +696,26 @@ init_fec (void) {
  * and then transforming it into a systematic matrix.
  */
 
-#define FEC_MAGIC	0xFECC0DEC
+enum { FEC_MAGIC = 0xFECC0DEC };
 
-void
-fec_free (fec_t *p) {
+zfex_status_code_t
+fec_free (fec_t *p)
+{
     assert (p != NULL && p->magic == (unsigned long)(((FEC_MAGIC ^ p->k) ^ p->n) ^ (uintptr_t) (p->enc_matrix)));
     free (p->enc_matrix);
     free (p);
+
+    return ZFEX_SC_OK;
 }
 
-fec_t *
-fec_new(unsigned short k, unsigned short n) {
+zfex_status_code_t
+fec_new(unsigned short k, unsigned short n, fec_t **out_fec_pp)
+{
+    if (out_fec_pp == NULL)
+    {
+        return ZFEX_SC_NULL_POINTER_INPUT;
+    }
+
     unsigned row, col;
     gf *p, *tmp_m;
 
@@ -603,16 +761,26 @@ fec_new(unsigned short k, unsigned short n) {
         *p = 1;
     free (tmp_m);
 
-    return retval;
+    *out_fec_pp = retval;
+
+    return ZFEX_SC_OK;
 }
 
 
-void
-fec_encode(const fec_t* code, const gf*restrict const*restrict const src, gf*restrict const*restrict const fecs, const unsigned*restrict const block_nums, size_t num_block_nums, size_t sz) {
-    unsigned char i, j;
-    size_t k;
-    unsigned fecnum;
-    const gf* p;
+zfex_status_code_t
+fec_encode(
+    fec_t const *code,
+    gf const * ZFEX_RESTRICT const * ZFEX_RESTRICT const src,
+    gf * ZFEX_RESTRICT const * ZFEX_RESTRICT const fecs,
+    const unsigned * ZFEX_RESTRICT const block_nums,
+    size_t const num_block_nums,
+    size_t const sz)
+{
+    unsigned int i = 0;
+    unsigned int j = 0;
+    size_t k = 0;
+    unsigned int fecnum = 0;
+    const gf* p = NULL;
 
     for (k = 0; k < sz; k += ZFEX_STRIDE)
     {
@@ -623,20 +791,24 @@ fec_encode(const fec_t* code, const gf*restrict const*restrict const src, gf*res
             fecnum = block_nums[i];
             assert(fecnum >= code->k);
             memset(fecs[i] + k, 0, stride);
+
             p = &(code->enc_matrix[fecnum * code->k]);
+
             for (j = 0; j < code->k; ++j)
             {
                 addmul(fecs[i] + k, src[j] + k, p[j], stride);
             }
         }
     }
+
+    return ZFEX_SC_OK;
 }
 
-int fec_encode_simd(
+zfex_status_code_t fec_encode_simd(
     fec_t const *code,
-    gf const *restrict const *restrict const inpkts,
-    gf *restrict const *restrict const fecs,
-    unsigned const *restrict const block_nums,
+    gf const * ZFEX_RESTRICT const * ZFEX_RESTRICT const inpkts,
+    gf * ZFEX_RESTRICT const * ZFEX_RESTRICT const fecs,
+    unsigned const * ZFEX_RESTRICT const block_nums,
     size_t const num_block_nums,
     size_t const sz)
 {
@@ -647,7 +819,7 @@ int fec_encode_simd(
     {
         if (((uintptr_t)inpkts[ix] % ZFEX_SIMD_ALIGNMENT) != 0)
         {
-            return EXIT_FAILURE;
+            return ZFEX_SC_BAD_INPUT_BLOCK_ALIGNMENT;
         }
     }
 
@@ -656,15 +828,15 @@ int fec_encode_simd(
     {
         if (((uintptr_t)fecs[ix] % ZFEX_SIMD_ALIGNMENT) != 0)
         {
-            return EXIT_FAILURE;
+            return ZFEX_SC_BAD_OUTPUT_BLOCK_ALIGNMENT;
         }
     }
 
-    unsigned i = 0;
-    unsigned j = 0;
+    unsigned int i = 0;
+    unsigned int j = 0;
     size_t k = 0;
-    unsigned fecnum = 0;
-    gf const *p;
+    unsigned int fecnum = 0;
+    gf const *p = NULL;
 
     for (k = 0; k < sz; k += ZFEX_STRIDE)
     {
@@ -685,7 +857,38 @@ int fec_encode_simd(
         }
     }
 
-    return EXIT_SUCCESS;
+    return ZFEX_SC_OK;
+}
+
+static zfex_status_code_t
+shuffle(gf const **pkt, unsigned int *index, unsigned int k)
+{
+    unsigned int i = 0;
+
+    for (i = 0; i < k; /* nop */)
+    {
+        if ((index[i] >= k) || (index[i] == i))
+        {
+            ++i;
+        }
+        else
+        {
+            /*
+             * put pkt in the right position (first check for conflicts).
+             */
+            unsigned int const c = index[i];
+
+            if (index[c] == c)
+            {
+                return ZFEX_SC_DECODE_INVALID_BLOCK_INDEX;
+            }
+#define SWAP(a, b, Tp) {Tp t = a; a = b; b = t;}
+            SWAP(index[i], index[c], unsigned int);
+            SWAP(pkt[i], pkt[c], gf const *);
+#undef SWAP
+        }
+    }
+    return ZFEX_SC_OK;
 }
 
 /**
@@ -694,7 +897,7 @@ int fec_encode_simd(
  * @param matrix a space allocated for a k by k matrix
  */
 void
-build_decode_matrix_into_space(const fec_t*restrict const code, const unsigned*const restrict index, const unsigned k, gf*restrict const matrix) {
+build_decode_matrix_into_space(const fec_t* ZFEX_RESTRICT const code, const unsigned*const ZFEX_RESTRICT index, const unsigned k, gf* ZFEX_RESTRICT const matrix) {
     unsigned char i;
     gf* p;
     for (i=0, p=matrix; i < k; i++, p += k) {
@@ -708,23 +911,43 @@ build_decode_matrix_into_space(const fec_t*restrict const code, const unsigned*c
     _invert_mat (matrix, k);
 }
 
-void
-fec_decode(const fec_t* code, const gf*restrict const*restrict const inpkts, gf*restrict const*restrict const outpkts, const unsigned*restrict const index, size_t sz) {
-    gf* m_dec = (gf*)alloca(code->k * code->k);
-    unsigned char outix=0;
-    unsigned char row=0;
-    unsigned char col=0;
+zfex_status_code_t
+fec_decode(
+    fec_t const *code,
+    gf const **inpkts,
+    gf * const *outpkts,
+    unsigned int *index,
+    size_t const sz)
+{
+    gf *m_dec = (gf *)alloca(code->k * code->k);
+    unsigned char outix = 0;
+    unsigned char row = 0;
+    unsigned char col = 0;
+
+    zfex_status_code_t const shuffle_sc = shuffle(inpkts, index, code->k);
+    if (shuffle_sc != ZFEX_SC_OK)
+    {
+        return shuffle_sc;
+    }
+
     build_decode_matrix_into_space(code, index, code->k, m_dec);
 
-    for (row=0; row<code->k; row++) {
-        assert ((index[row] >= code->k) || (index[row] == row)); /* If the block whose number is i is present, then it is required to be in the i'th element. */
-        if (index[row] >= code->k) {
+    for (row = 0; row < code->k; ++row)
+    {
+        assert((index[row] >= code->k) || (index[row] == row)); /* If the block whose number is i is present, then it is required to be in the i'th element. */
+
+        if (index[row] >= code->k)
+        {
             memset(outpkts[outix], 0, sz);
-            for (col=0; col < code->k; col++)
+            for (col = 0; col < code->k; ++col)
+            {
                 addmul(outpkts[outix], inpkts[col], m_dec[row * code->k + col], sz);
-            outix++;
+            }
+            ++outix;
         }
     }
+
+    return ZFEX_SC_OK;
 }
 
 /**
